/******************************************************************************
*******************************************************************************
******                                                                  *******
******     This code is written by Abdullah Al Mueen at the department  *******
******     of Computer Science and Engineering of University of         *******
******     California - RIverside.                                      *******
******                                                                  *******
*******************************************************************************
******************************************************************************/

/*#############################################################################
######                                                                  #######
######     This code is open to use, reuse and distribute at the user's #######
######     own risk and as long as the above credit is ON. The complete #######
######     description of the algorithm and methods applied can be      #######
######     found in the paper                                           #######
######                                                                  #######
#############################################################################*/

/******************************************************************
* This code are been adapted for Linux and MAC OS by :
* Thomas Guyet and René Quiniou, INRIA-Centre de Rennes, France.
******************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "mex.h"

#ifdef WIN32
    #include <window.h>
    #include <conio.h>
    #include <process.h> 
    #include <signal.h>
#else
    #include <pthread.h>
    #include <unistd.h>
#endif

#define INF 999999999999.0

//internal data structure
double **data;
long long tickCounter = 0;
double *r;
double *buffer;

#ifdef WIN32
    HANDLE hThread;
#else
    pthread_t thread;
#endif
unsigned int threadID = 0;

//output variables
long long loc1 , loc2 , i;
double bsf;

//input variables
long long DBSIZE;
long long length;
long long LENGTH;
long long W;
double slp;
int ZZ = 0;

//structures
struct seqNode;

struct neighborNode{
    struct seqNode *seq;
    struct neighborNode *next;
    long long tStamp;
    double d;
};

struct seqNode{
    long long tStamp;
    double * seq;

    struct seqNode *next;

    struct neighborNode * neighborList;

    double dR;
    struct seqNode *nextInOrder;
    struct seqNode *prevInOrder;

    struct neighborNode * rNeighborList;
    long long NNCount;
};

struct seqNode * seqStart, * orderStart, *orderEnd;

double distance(double *x, double *y, long long length , double best_so_far = INF )
{
    long long i;
    double sum = 0;
    double bsf2 = best_so_far*best_so_far;
    for ( i = 0 ; i <= length && sum < bsf2 ; i++ )
        sum += (x[i]-y[i])*(x[i]-y[i]);
    return sqrt(sum);
}

void error(int id)
{
    if(id==1)
        printf("ERROR : Memory can't be allocated!!!\n\n");
    else if ( id == 2 )
        printf("ERROR : File not Found!!!\n\n");
    else if ( id == 3 )
        printf("ERROR : Can't create Output File!!!\n\n");
    else if ( id == 4 )
        printf("ERROR : Invalid Number of Arguments!!!\n\n");
    else if ( id == 5 )
        printf("ERROR : Impossible to create thread!!!\n\n");
    system("PAUSE");
    exit(1);
}

double findUpperBound(seqNode *curSeqNode , long long t)
{
    struct neighborNode * n = curSeqNode->neighborList;
    double d = INF;
    
    if( n == NULL )
        return INF;
    else {
        while(n!=NULL) {
            if(n->tStamp>=t)
                d = n->d;
            else
                break;
            n = n->next;
        }
    }
    return d;
}

double insertNeighbor(seqNode *curSeqNode , seqNode* s, double d)
{
    struct neighborNode * n = curSeqNode->neighborList , *p;

    if( n == NULL ) {
        n = new struct neighborNode;
        n->d = d;
        n->next = NULL;
        n->seq = s;
        n->tStamp = s->tStamp;
        curSeqNode->neighborList = n;
        curSeqNode->NNCount++;
    } else {
        while( n->next != NULL ) {
            if( d > n->d ) {
                if( s->tStamp > n->tStamp ) {
                    struct neighborNode *nn = n->next;
                    n->next = new struct neighborNode;
                    n->next->next = nn;

                    n->next->d = n->d;
                    n->d = d;

                    n->next->tStamp = n->tStamp;
                    n->tStamp = s->tStamp;

                    n->next->seq = n->seq;
                    n->seq = s;
                    curSeqNode->NNCount++;
                }
                break;
                
            } else { // it means d <= n->d
                if( s->tStamp > n->tStamp ) {
                    struct neighborNode *nn = n->next;

                    if( n == curSeqNode->neighborList )
                         curSeqNode->neighborList = n->next;
                    else
                         p->next = n->next;
                         
                    delete n;
                    n = nn;
                    curSeqNode->NNCount--;
                } else {
                    p = n;
                    n=n->next;
                }
            }
         }//end while(n->next!=NULL)
        if( n->next == NULL ) {   //if at the end of the list
            if( d > n->d ) {
                if( s->tStamp > n->tStamp ) {
                    n->next = new struct neighborNode;
                    n->next->next = NULL;

                    n->next->d = n->d;
                    n->d = d;

                    n->next->tStamp = n->tStamp;
                    n->tStamp = s->tStamp;

                    n->next->seq = n->seq;
                    n->seq = s;
                     curSeqNode->NNCount++;
                }
            } else { // it means d <= n->d
                if( s->tStamp > n->tStamp ) {
                    n->d = d;
                    n->seq = s;
                    n->tStamp = s->tStamp;
                    n->next = NULL;

                } else {
                    n->next = new struct neighborNode;
                    n->next->next = NULL;
                    n->next->d = d;
                    n->next->tStamp = s->tStamp;
                    n->next->seq = s;
                     curSeqNode->NNCount++;
                }
            }
        }//end if ( n->next == NULL )

     }//end else of if(n == NULL)

    if( curSeqNode->tStamp - curSeqNode->neighborList->tStamp <= LENGTH )
        return curSeqNode->neighborList->d;
    else
        return INF;
}


double insertReverseNN(seqNode *curSeqNode , seqNode* s, double d)
{
    struct neighborNode * n = curSeqNode->rNeighborList;

    if( n == NULL ) {
        n = new struct neighborNode;
        n->d = d;
        n->next = NULL;
        n->seq = s;
        n->tStamp = s->tStamp;
        curSeqNode->rNeighborList = n;
    } else {
        struct neighborNode * nn = n->next;
        n->next = new struct neighborNode;
        n = n->next;
        n->d = d;
        n->next = nn;
        n->seq = s;
        n->tStamp = s->tStamp;
    }//end else of if(n == NULL)

    return curSeqNode->rNeighborList->d;
}


void buildNNList(seqNode *curSeqNode)
{
    double curBSF = INF;
    double ub = INF;

    struct seqNode *s = curSeqNode->nextInOrder;
    struct seqNode *t = curSeqNode->prevInOrder;
    curSeqNode->neighborList = NULL;


    int stop = 0;

    while( !stop ) {
        stop = 1;
        if(s != NULL && fabs(s->dR-curSeqNode->dR) < curBSF) {
            if( ( curSeqNode->tStamp - s->tStamp ) >= LENGTH && ( curSeqNode->tStamp - s->tStamp ) <= W-LENGTH+1 ) {
                ub = findUpperBound(curSeqNode , s->tStamp);
                double d = distance(s->seq,curSeqNode->seq,LENGTH,ub);
                if( d < ub )
                    curBSF = insertNeighbor(curSeqNode, s, d);
            }
            s = s->nextInOrder;
            stop = 0;
        }

        if(t != NULL && fabs(t->dR-curSeqNode->dR) < curBSF) {
            if( ( curSeqNode->tStamp - t->tStamp ) >= LENGTH && ( curSeqNode->tStamp - t->tStamp ) <= W-LENGTH+1 ) {
                ub = findUpperBound(curSeqNode , t->tStamp);
                double d = distance(t->seq,curSeqNode->seq,LENGTH,ub);
                if( d < ub )
                    curBSF = insertNeighbor(curSeqNode, t, d);
            }
            t = t->prevInOrder;
            stop = 0;
        }
    }//end while(s!=NULL)

    //Reverse the link list
    struct neighborNode *n = curSeqNode->neighborList;
    if( n != NULL ) {
        struct neighborNode * p = n->next;
        if( p != NULL ) {
            n->next = NULL;
            struct neighborNode * m = p->next;
            while( m!=NULL ) {
               p->next = n;
               n = p;
               p = m;
               m = m->next;
            }
            p->next = n;
            curSeqNode->neighborList = p;
        }
        if( curSeqNode->neighborList->d < bsf ) {
            loc1 = curSeqNode->neighborList->seq->tStamp;
            loc2 = curSeqNode->tStamp;
            bsf = curSeqNode->neighborList->d;
        }

        //update the Reverse Nearest Neighbor List
        insertReverseNN(curSeqNode->neighborList->seq,curSeqNode,curSeqNode->neighborList->d);
    }//end if n != NULL
}

#ifdef WIN32
unsigned int mainThread(void * pVoid)
#else
void *mainThread(void * pVoid)
#endif
{
    double d;
    long long j ;
    double ex , ex2 , mean, std;

    double t1,t2;
    double count = 0;
    double t11,t22,wcase = 0.0;
    struct seqNode * curSeqNode = NULL, *nodeToDelete;
    
    tickCounter = 0;

    bsf = INF;
    loc1 = -1;
    loc2 = -1;

    i = 0;
    j = 0;
    ex = ex2 = 0;

    t1 = clock();

    srand(time(NULL));
    
    DBSIZE = length-LENGTH+1;

    data = (double **)malloc(sizeof(double *)*DBSIZE);
    if( data == NULL )
        error(1);

    r = (double *)malloc(sizeof(double)*LENGTH);

    //This is the loop that itearates over each clock tick
    while( i < length ) {
        d = buffer[i];  
        //advance the clock tick Counter
        
        t11 = clock();
        
        if( i < DBSIZE ) {
            data[i] = (double *)malloc(sizeof(double)*LENGTH);
            if( data[i] == NULL )
                error(1);
        }

        ex += d;
        ex2 += d*d;

        long long k = LENGTH-1;
        for(j=((i>=DBSIZE)?DBSIZE-1:i);j>=((i>=LENGTH)?(i-k):0);j--)
            data[j][i-j] = d;

         //Handle Deletion
        if( i >= W ) {
            nodeToDelete = seqStart;
            seqStart = seqStart->next;

            if( nodeToDelete == orderStart ) {
                orderStart->nextInOrder->prevInOrder = orderStart->prevInOrder; //should be NULL
                orderStart = orderStart->nextInOrder;
            } else if( nodeToDelete == orderEnd ) {
                orderEnd->prevInOrder->nextInOrder = orderEnd->nextInOrder;
                orderEnd = orderEnd->prevInOrder;
            }  else {
                nodeToDelete->prevInOrder->nextInOrder = nodeToDelete->nextInOrder;
                nodeToDelete->nextInOrder->prevInOrder = nodeToDelete->prevInOrder;
            }
            nodeToDelete->nextInOrder = NULL;
            nodeToDelete->prevInOrder = NULL;
            nodeToDelete->next = NULL;
        }

        if( i >= k ) {
            tickCounter ++;
            //z-Normalize the seq Node created LENGTH ticks before
            mean = ex/LENGTH;
            std = ex2/LENGTH;
            std = sqrt(std-mean*mean);
            ex -= data[i-k][0];
            ex2 -= data[i-k][0]*data[i-k][0];
            if( ZZ == 1 ) {
                for( long long u = 0 ; u < LENGTH ; u++ )
                    data[i-k][u] = (data[i-k][u]-mean)/std;
            }
            //Create a seq Node for this subsequence
            if( curSeqNode == NULL ) {
                //first sequence node.
                orderEnd = orderStart = seqStart = curSeqNode = new struct seqNode;
                curSeqNode->tStamp = tickCounter;
                curSeqNode->next = NULL;
                curSeqNode->seq = data[i-k];
                curSeqNode->neighborList = NULL;
                curSeqNode->NNCount = 0;
                curSeqNode->rNeighborList = NULL;
                curSeqNode->dR = 0;
                curSeqNode->nextInOrder = curSeqNode->prevInOrder = NULL;//curSeqNode;
                for(long long u = 0 ; u < LENGTH ; u++ )
                    r[u] = data[i-k][u];
            } else {
                //subsequent sequence nodes
                curSeqNode->next = new struct seqNode;
                curSeqNode = curSeqNode->next;
                curSeqNode->tStamp = tickCounter;
                curSeqNode->next = NULL;
                curSeqNode->seq = data[i-k];
                curSeqNode->dR = distance(r,data[i-k],LENGTH);
                curSeqNode->rNeighborList = NULL;
                curSeqNode->NNCount = 0;


                //insert into order line which is a doubly linked list implemented through the nextInOrder and prevInOrder
                
                struct seqNode * s = orderStart;
                while(s!=NULL) {
                    if( curSeqNode->dR >= s->dR )
                        s = s->nextInOrder;
                    else {
                        curSeqNode->nextInOrder = s;
                        curSeqNode->prevInOrder = s->prevInOrder;
                        if( s->prevInOrder != NULL )
                            s->prevInOrder->nextInOrder = curSeqNode;
                        s->prevInOrder = curSeqNode;
                        if( s == orderStart )
                            orderStart = curSeqNode;
                        break;
                    }//end else

                }//end while(s!=NULL)
                if( s == NULL ) {
                    orderEnd->nextInOrder = curSeqNode;
                    curSeqNode->nextInOrder = NULL;
                    curSeqNode->prevInOrder = orderEnd;
                    orderEnd = curSeqNode;
                }

                //handle Insertions
                buildNNList(curSeqNode);

            }//end else of if(curSeqNode==NULL)
         }//end else of if i >= k


         if( i >= W ) {
            struct neighborNode * n = nodeToDelete->rNeighborList;
            
            while(n != NULL) {
                struct seqNode * s = n->seq;
                struct neighborNode *m , *t ;

                t = s->neighborList;
                m = s->neighborList = t->next;

                if(m!=NULL)
                    insertReverseNN(m->seq,s,m->d);

                delete t;

                t = n;
                n = n->next;
                delete t;
            }


            if( nodeToDelete->tStamp == loc1 ) {
                //Update the bsf
                bsf = INF;
                struct seqNode * s = seqStart;
                while( s != NULL )  {
                    if( s->neighborList != NULL && s->neighborList->d < bsf && s->tStamp > nodeToDelete->tStamp && s->neighborList->seq->tStamp > nodeToDelete->tStamp ) {
                        loc1 = s->neighborList->seq->tStamp;
                        loc2 = s->tStamp;
                        bsf = s->neighborList->d;
                    }
                    s = s->next;
                }
            }

            free(data[i-W]);
            delete nodeToDelete;
        }//if i>=w

        i++;
        t22 = clock();
        if( (t22-t11)/CLOCKS_PER_SEC < slp ) {
#ifdef WIN32
            Sleep(1000*(slp-(t22-t11)/CLOCKS_PER_SEC));
#else
            sleep((slp-(t22-t11)/CLOCKS_PER_SEC));
#endif
        }
    }//end while(fscanf////....

    t2 = clock();
#ifdef WIN32
    return 1;
#else
    return NULL;
#endif
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    double * ddd;

    if (nrhs == 0) {
         //TerminateThread(hThread);
        length = 0;

#ifdef WIN32
        Sleep(1000);
        
        if( threadID != 0 ) {
            while(CloseHandle(hThread)==0) {
                    mexPrintf("can not close\n");
            }
            threadID = 0;
        }
#else
        sleep(1);
        if( threadID != 0 ) {
            while(pthread_cancel(thread)!=0) {
                    mexPrintf("can not close\n");
            }
            threadID = 0;
        }
#endif
        return;
    } else if(nrhs == 1) {
        plhs[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
        double * outArray = mxGetPr(plhs[0]);
        outArray[0] = bsf;

        plhs[1] = mxCreateDoubleMatrix(1, 1, mxREAL);
        outArray = mxGetPr(plhs[1]);
        outArray[0] = loc1+1;

        plhs[2] = mxCreateDoubleMatrix(1, 1, mxREAL);
        outArray = mxGetPr(plhs[2]);
        outArray[0] = loc2+1;


        plhs[3] = mxCreateDoubleMatrix(1, 1, mxREAL);
        outArray = mxGetPr(plhs[3]);
        outArray[0] =  i+1;
        return;
    } else {
        const mxArray *bData = prhs[0];
        ddd = mxGetPr(bData);

        const mxArray *lData = prhs[1];
        length = mxGetScalar(lData);

        buffer = (double *)malloc(sizeof(double)*length);
        memcpy(buffer,ddd,sizeof(double)*length);

        const mxArray *tData = prhs[2];
        LENGTH = mxGetScalar(tData);

        const mxArray *wData = prhs[3];
        W = mxGetScalar(wData);
    
        const mxArray *sData = prhs[4];    
        slp = mxGetScalar(sData);

        ZZ = 1;

#ifdef WIN32
        hThread = (HANDLE)_beginthreadex( NULL, 0, (unsigned int (__stdcall *)(void *))&mainThread, NULL, 0, &threadID );
#else
        int rc = pthread_create(&thread, NULL, mainThread, (void *)threadID);
        if(rc) error(5);
#endif
    }
}


